import io
import unittest
import zipfile
from pathlib import Path
from typing import Optional, NamedTuple

from pynanto import StringResource, PathResource
from pynanto.common.bundle import bundle_definition, default_item_filter, build_archive

parent = Path(__file__).parent


class Test_bundle_definition(unittest.TestCase):
    support_data = parent / 'support_data/bundle_definition'

    def test_one_file(self):
        folder = self.support_data / 'one_file'
        actual = set(bundle_definition(folder))
        expect = set([PathResource('foo.py', folder / 'foo.py')])
        self.assertEqual(expect, actual)

    def test_zero_file(self):
        folder = self.support_data / 'zero_file'
        folder.mkdir(exist_ok=True)  # git does not commit empty folders
        actual = set(bundle_definition(folder))
        expect = set([])
        self.assertEqual(expect, actual)

    def test_selective(self):
        folder = self.support_data / 'relative_to'
        actual = set(bundle_definition(folder / 'yes', relative_to=folder))
        expect = set([PathResource('yes/yes.txt', folder / 'yes/yes.txt')])
        self.assertEqual(expect, actual)

    def test_item_filter(self):
        folder = self.support_data / 'item_filter'
        reject = folder / 'yes/reject'

        pycache = folder / 'yes/__pycache__'
        pycache.mkdir(exist_ok=True)
        (pycache / 'cache.txt').write_text('some cache')

        def item_filter(item: PathResource) -> Optional[PathResource]:
            if item.filepath == reject:
                return None
            return default_item_filter(item)

        actual = set(bundle_definition(folder, item_filter=item_filter))
        expect = set([PathResource('yes/yes.txt', folder / 'yes/yes.txt')])
        self.assertEqual(expect, actual)


class Test_build_archive(unittest.TestCase):
    support_data = parent / 'support_data/build_archive'

    def test_simple(self):
        class ZFile(NamedTuple):
            filename: str
            content: bytes

        folder = self.support_data / 'simple'
        (folder / 'empty_dir').mkdir(exist_ok=True)  # should be ignored by build_archive

        archive_bytes = build_archive(list(bundle_definition(folder)) +
                                      [StringResource('dir1/baz.txt', "#baz")])

        actual_files = set()
        with zipfile.ZipFile(io.BytesIO(archive_bytes)) as zf:
            for il in zf.infolist():
                with zf.open(il, 'r') as file:
                    actual_files.add(ZFile(il.filename, file.read()))

        expected_files = {
            ZFile('foo.txt', '#foo'.encode()),
            ZFile('dir1/bar.txt', '#bar'.encode()),
            ZFile('dir1/baz.txt', '#baz'.encode()),
        }

        self.assertEqual(expected_files, actual_files)


if __name__ == '__main__':
    unittest.main()



from threading import Thread
from typing import Optional

import uvicorn
from fastapi import FastAPI
from fastapi.responses import Response

import pynanto
from pynanto import Webserver
from pynanto.routes import Route


class WsFastapi(Webserver):
    def __init__(self):
        super().__init__()
        self.app = FastAPI()
        self.thread: Optional[Thread] = None

    def _setup_route(self, route: Route):
        def func(*args) -> Response:
            return self.to_native_response(route.callback())

        self.app.add_route(route.path, route=func)

    def to_native_response(self, pn_response: pynanto.Response):
        return Response(content=pn_response.content, media_type=pn_response.content_type)

    def _start_listen(self):
        def run():
            uvicorn.run(self.app, host=self.host, port=self.port)

        self.thread = Thread(target=run, daemon=True)
        self.thread.start()



from pynanto import Response, Routes, Webserver
from pynanto.response import Request
from pynanto.server import find_port
from pynanto.server.fetch import sync_fetch_response
from tests import for_all_webservers


@for_all_webservers()
def test_webservers_get(webserver: Webserver):
    response_a = Response('a', 'text/plain')
    response_b = Response('b', 'text/html')

    routes = (
        Routes()
        .add_route('/b', lambda req: response_b)
        .add_route('/', lambda req: response_a)
    )

    webserver.set_routes(routes).set_port(find_port()).start_listen().wait_ready()

    url = webserver.localhost_url()

    assert sync_fetch_response(url) == response_a
    assert sync_fetch_response(url + '/b') == response_b


@for_all_webservers()
def test_webservers_post(webserver: Webserver):
    # GIVEN
    response_a = Response('a', 'text/plain')
    actual_request: Request = None

    def handler(req: Request) -> Response:
        nonlocal actual_request
        actual_request = req
        return response_a

    routes = Routes().add_route('/rpc', handler)

    webserver.set_routes(routes).set_port(find_port()).start_listen().wait_ready()

    url = webserver.localhost_url()

    # WHEN
    actual_response = sync_fetch_response(url + '/rpc', method='POST', data='post-body')

    # THEN
    assert actual_response == response_a
    assert actual_request.method == 'POST'
    assert actual_request.content.decode('utf8') == 'post-body'



from pathlib import Path

from pynanto.bundles import external_filename, path_is_contained


def test_external_filename():
    actual = external_filename()
    assert actual == Path(__file__).resolve()


def test_is_contained_into():
    assert path_is_contained(Path('/foo/bar'), Path('/foo/bar/baz')) is False


def test_is_contained_into_2():
    assert path_is_contained(Path('/foo/xxx'), Path('/foo/bar/baz')) is False


def test_is_contained_into_3():
    assert path_is_contained(Path('/foo/bar/baz/yyy'), Path('/foo/bar/baz')) is True


def test_is_contained_into_4():
    assert path_is_contained(Path('/foo/bar/baz'), Path('/foo/bar/baz')) is True



import zipfile
from io import BytesIO
from pathlib import Path
from typing import Iterator, Optional, Callable

from pynanto.resource import PathResource
from pynanto.resource import Resource, StringResource


def default_item_filter(item: Resource) -> Optional[Resource]:
    if not isinstance(item, PathResource):
        return item
    item: PathResource
    filepath = item.filepath
    if filepath.name == '.DS_Store':
        return None
    if filepath.name == '__pycache__' and filepath.is_dir():
        return None
    return item


def bundle_definition(
        folder: Path, relative_to: Optional[Path] = None,
        item_filter: Callable[[PathResource], Optional[PathResource]] = default_item_filter
) -> Iterator[PathResource]:
    if relative_to is None:
        relative_to = folder

    def recurse(path: Path):
        for f in path.glob('*'):
            rel = f.relative_to(relative_to)
            candidate = PathResource(str(rel), f)
            item = item_filter(candidate)
            if item is not None:
                if f.is_file():
                    yield item
                else:
                    yield from recurse(f)

    yield from recurse(folder)
    return iter(())


def build_archive(item_iterator: Iterator[Resource]) -> bytes:
    stream = BytesIO()
    zf = zipfile.ZipFile(stream, "w", compression=zipfile.ZIP_DEFLATED, compresslevel=1)

    for item in item_iterator:
        if isinstance(item, PathResource):
            zf.write(item.filepath, item.arcname)
        elif isinstance(item, StringResource):
            zf.writestr(item.arcname, item.get_content())
        else:
            raise Exception(f'Unhandled class \n  type={type(item).__name__} \n  data={item}')
    zf.close()

    stream.seek(0)
    return stream.getbuffer().tobytes()



import unittest

from js import document, HTMLElement
from pynanto.remote.unittest_fix import unittest_main_fixed
from pynanto.remote.widget import Widget


class WidgetTestCase(unittest.TestCase):
    def test_basic_bind(self):
        class Widget1(Widget):
            def __init__(self):
                super().__init__("<span id='span1'>foo</span>")
                self.span1: HTMLElement = self
                self.bind_self_elements()

        target = Widget1()
        self.assertEqual('foo', target.span1.innerHTML)

    def test_bind_to_widget(self):
        class WidgetSub(Widget):
            def __init__(self):
                super().__init__("<div>I'm WidgetSub</div>")

        class Widget1(Widget):
            def __init__(self):
                super().__init__("<div id='div1'></div>")
                self.div1 = self(lambda: WidgetSub())

        target = Widget1()
        target.append_to(document.createElement('div'))
        html = target.container.innerHTML
        self.assertTrue("I'm WidgetSub" in html, f'Actual html=```{html}```')

    def test_bind_events(self):
        actual = []

        class Widget1(Widget):
            def __init__(self):
                super().__init__("<button id='foo'>foo</button>")

            def foo__click(self, *args):
                actual.append(1)

        target = Widget1()
        foo = target.container.querySelector('#foo')
        foo.click()
        self.assertEqual([1], actual)


def main():
    unittest_main_fixed(__name__)
    # unittest.main()


if __name__ == '__main__':
    main()



import sys
import traceback
from pathlib import Path
from types import ModuleType
from typing import Optional, Iterator

from pynanto.bootstrap import Bootstrap
from pynanto.bundles import Bundles, external_filename
from pynanto.resource import Resource, PathResource
from pynanto.response import Response, Request
from pynanto.routes import Routes
from pynanto.rpc import Services, Stubber, Module
from pynanto.webserver import Webserver
from pynanto.webserver import wait_forever
from pynanto.webservers.available_webservers import available_webservers


class Config:
    def __init__(self):
        self._routes = Routes()
        self.bootstrap = Bootstrap()
        self.bundles = Bundles()
        self._webserver: Optional[Webserver] = None
        self.bundle_route = '/pynanto/bundle/default.zip'

    @property
    def webserver(self) -> Webserver:
        if self._webserver is None:
            self._webserver = available_webservers().new_instance()
        return self._webserver

    def set_routes(self, routes: Routes) -> 'Config':
        if not self._routes.is_empty:
            raise Exception('routes are already set')
        self._error_if_attached()
        self._routes = routes
        return self

    def quickstart(
            self,
            blocking: bool = True,
            webserver_instance: Optional[Webserver] = None,
            port: Optional[int] = None,
            stack_backtrack=1,
            rpc_module: ModuleType = None
    ) -> 'Config':
        if rpc_module is None:
            try:
                import server.rpc
                rpc_module = server.rpc
            except:
                traceback.print_exc()

        services = Services()
        if rpc_module is not None:
            rpc_module = Module(rpc_module)
            services.add_module(rpc_module)

        self.set_routes(
            Routes()
            .add_route(self.bundle_route, self.bundles.to_response)
            .add_route('/', self.quickstart_index_response)
            .add_route_obj(services.route)
        )

        if webserver_instance is None:
            webserver_instance = available_webservers().new_instance()
        if port is not None:
            webserver_instance.set_port(port)
        rpc_url = services.route.path
        stubber = Stubber(rpc_url, services, rpc_module)

        # one should be able to import:
        # - through an 'import module_name'
        # - through ??

        ef = external_filename(stack_backtrack)
        print(f'detected user dir: {ef}')
        print(f'{sys.path}')
        if ef is not None:
            self.bundles.add_flat_folder(ef.parent / 'app', relative_to=ef.parent)
            self.bundles.add_flat_folder(ef.parent / 'remote', relative_to=ef.parent)
            self.bundles.add_flat_folder(ef.parent / 'common', relative_to=ef.parent)
            remotepy = ef.parent / 'remote.py'

            def gen_remotepy() -> Iterator[Resource]:
                if remotepy.exists():
                    yield PathResource('remote.py', remotepy)
                yield from stubber.remote_stub_resources()

            self.bundles.add_resources(gen_remotepy)

        pynanto_remote = Path(__file__).parent
        self.bundles.add_flat_folder(pynanto_remote, relative_to=pynanto_remote.parent)
        # language=python
        load_default_bundle = f"""
import sys
from pyodide.http import pyfetch
response = await pyfetch('{self.bundle_route}')
await response.unpack_archive(extract_dir='/default_bundle')
sys.path.insert(0, '/default_bundle')
"""
        self.bootstrap.add_python(load_default_bundle)
        # language=python
        self.bootstrap.add_python("""
import remote
await remote.main()
""")

        self.attach_webserver(webserver_instance)
        self.webserver.start_listen()

        if blocking:
            wait_forever()
        else:
            self.webserver.wait_ready()

        return self

    def quickstart_index_response(self, request: Request) -> Response:
        bootstrap_javascript = self.bootstrap.get_javascript()
        return Response(
            f'<h1>loading...</h1>'
            f'<script>{bootstrap_javascript}</script>'
            , 'text/html'
        )

    def attach_webserver(self, webserver: Webserver) -> 'Config':
        self._error_if_attached()
        self._webserver = webserver
        webserver.set_routes(self._routes)
        return self

    def _error_if_attached(self):
        if self._webserver is not None:
            raise Exception('Webserver already attached')



def forward_to(host, guest_name: str):
    def ga(self, attr):
        guest_obj = getattr(self, guest_name)
        return getattr(guest_obj, attr)

    def sa(self, key, value):
        if key == guest_name:
            super(host, self).__setattr__(key, value)
        else:
            guest_obj = getattr(self, guest_name)
            setattr(guest_obj, key, value)

    host.__getattr__ = ga
    host.__setattr__ = sa

    return host


### tests

class Guest:
    def __init__(self, string: str):
        self.string = string

    def foo(self):
        return self.string


class Host1:
    def __init__(self, guest):
        self.guest = guest
        self.prop1 = "prop1-val"

    def bar(self):
        return self.prop1


def test_proxy():
    class Host2:
        def __init__(self, guest):
            self._guest = guest

    forward_to(Host1, 'guest')
    target1 = Host1(Guest("foo"))
    assert target1.foo() == "foo"

    forward_to(Host2, '_guest')
    target2 = Host2(Guest("bar"))
    assert target2.foo() == "bar"


def test_non_existant():
    forward_to(Host1, 'guest')
    target1 = Host1(Guest("foo"))
    try:
        target1.does_not_exist()
        exception = None
    except Exception as ex:
        exception = ex

    assert exception is not None


def test_shouldNotForward_existingHostMethods():
    forward_to(Host1, 'guest')
    target1 = Host1(Guest("foo"))
    assert "prop1-val" == target1.bar()
    assert "prop1-val" == target1.prop1


def test_shouldForwardSet():
    forward_to(Host1, 'guest')
    guest = Guest("foo")
    target1 = Host1(guest)

    target1.guestprop1 = "hello"
    assert "hello" == guest.guestprop1


def test_HostSetter():
    class HostLoc:
        def __init__(self, guest):
            self.guest = guest
            self.prop1 = "prop1-val"

        def reset(self):
            self.guest.string = self.prop1

    forward_to(HostLoc, 'guest')

    guest = Guest("init")
    target = HostLoc(guest)
    target.prop1 = "changed1"
    target.reset()
    assert "changed1" == target.string
    # assert not hasattr(guest, 'prop1')



import urllib
import urllib.request

from pynanto import Response


async def async_fetch_str(url: str, method: str = 'GET', data: str = '') -> str:
    response = sync_fetch_response(url, method=method, data=data)
    return response.content


def sync_fetch_response(url: str, method: str = 'GET', data: str = '') -> Response:
    def make_response(r):
        return Response(
            r.read().decode("utf-8"),
            r.headers.get_content_type()
        )

    if method != 'GET':
        if isinstance(data, str):
            data = bytes(data, 'utf8')
        rq = urllib.request.Request(url, method=method, data=data)
        with urllib.request.urlopen(rq) as r:
            return make_response(r)
    else:
        with urllib.request.urlopen(url) as r:
            return make_response(r)



# Browse pyscript virtual filesystem
from __future__ import annotations

import os
from pathlib import Path
from typing import TypeVar

from pyodide.ffi import create_proxy

from js import console

HTMLElement = TypeVar('HTMLElement')
# from app.browser.html.dom_definitions import HTMLElement
# from app.browser.html.dom_helpers import download_file
from pynanto.remote.widget import Widget


# from app.common.filesystem import zip_in_memory


class FilesystemTreeWidget(Widget):
    def __init__(self, path: Path = Path('/'), indent=1):
        self._indent = indent
        super().__init__(  # language=HTML
            f"""
            <span id="_entity"></span>
            <span id="_download">&nbsp ↓ &nbsp</span>
            <div id="_children" style="margin-left: 1em"></div>
            
            """
        )
        self.path = Path(path)
        self._entity: HTMLElement = self
        self._children: HTMLElement = self
        self._download: HTMLElement = self

    def after_render(self):

        self._entity.onclick = create_proxy(self.toggle_display)
        self._download.onclick = create_proxy(self.download)

        # if not self.is_dir():
        #     self._download.style.display = 'none'

        if self._indent > 1:
            self.toggle_display()

        self._update_caption()

        if self.is_dir():
            self._recurse()

    def is_dir(self):
        # return self.path.is_dir() # in pyodide: `PermissionError: [Errno 63] Operation not permitted: '/proc/self/fd'`
        return os.path.isdir(self.path)

    def toggle_display(self, *args):
        self._children.style.display = '' if self._children_hidden() else 'none'
        self._update_caption()

    def download(self, *args):
        console.log(f'click {self.path.absolute()}')
        if self.is_dir():
            download_file(self.path.name + '.zip', zip_in_memory(self.path))
        else:
            download_file(self.path.name, self.path.read_bytes())

    def _update_caption(self):
        if_dir = '▸' if self._children_hidden() else '▼'
        pre = if_dir if self.is_dir() else ' '
        self._entity.innerHTML = (f'<span style="display:inline-block; width: 1em">{pre}</span>'
                                  + ' ' + self.path.name)

    def _children_hidden(self):
        s = self._children.style
        return not s.display == ''

    def _recurse(self):
        for child in self.path.glob('*'):
            w = FilesystemTreeWidget(child, self._indent + 1)
            w.append_to(self._children)



class PynantoException(Exception):
    pass


class RemoteException(PynantoException):
    """When user defined server code fails"""
    pass


class RemoteError(PynantoException):
    """When network/pynanto infrastructure fails"""
    pass



from playwright.sync_api import Page, expect

import pynanto
from pynanto import Bootstrap, Response, Routes, Webserver
from pynanto.server import find_port
from tests import for_all_webservers


@for_all_webservers()
def test_bootstrap(page: Page, webserver: Webserver):
    bootstrap_javascript = Bootstrap().add_python(
        'from js import document\n' +
        'document.getElementById("tag1").value = "Hello world!"\n'
    ).get_javascript()

    response = Response(f'<input id="tag1" value="Hello">'
                        f'<script>{bootstrap_javascript}</script>', 'text/html')

    routes = Routes().add_route('/', lambda _: response)
    webserver.set_routes(routes).set_port(find_port()).start_listen().wait_ready()

    page.goto(webserver.localhost_url())

    expect(page.locator('id=tag1')).to_have_value('Hello world!')


@for_all_webservers()
def test_bootstrap_config(page: Page, webserver: Webserver):
    cfg = pynanto.Config()
    cfg.bootstrap.add_python(
        'from js import document\n' +
        'document.getElementById("tag1").innerHTML = "Hello bootstrap!"\n'
    )

    cfg.set_routes(Routes().add_route(
        '/',
        lambda _: Response(f'<div id="tag1"></div>'
                           f'<script>{cfg.bootstrap.get_javascript()}</script>', 'text/html')
    ))

    cfg.attach_webserver(webserver)
    webserver.set_port(find_port()).start_listen().wait_ready()

    page.goto(webserver.localhost_url())
    expect(page.locator('id=tag1')).to_have_text('Hello bootstrap!')



import urllib.error
import urllib.request
from time import sleep


def wait_url(url: str):
    for _ in range(300):
        try:
            urllib.request.urlopen(url)
            return
        except urllib.error.HTTPError:
            return
        except Exception:
            sleep(0.01)
    raise Exception(f'timeout wait_url(`{url}`)')
